@@ -1,5 +1,6 @@
 require 'csv'
 require 'puppet'
+require 'timeout'
 
 class Puppet::Error::OpenstackAuthInputError < Puppet::Error
 end
@@ -10,7 +11,49 @@ end
 class Puppet::Provider::Openstack < Puppet::Provider
 
   initvars # so commands will work
-  commands :openstack => 'openstack'
+  commands :openstack_command => 'openstack'
+
+  # this actions are not idempotent and retries can cause
+  # duplications or endless loops
+  def self.no_retry_actions
+    %w(create remove delete)
+  end
+
+  # timeout the openstack command
+  # after this number of seconds
+  # retry the command until the request_timeout
+  def self.command_timeout
+    20
+  end
+
+  # timeout the entire request with error
+  # after this number of seconds
+  def self.request_timeout
+    60
+  end
+
+  # sleep for this number of seconds
+  # between command retries
+  def self.retry_sleep
+    3
+  end
+
+  # run the openstack command
+  # with command_timeout
+  def self.openstack(*args)
+    begin
+      Timeout.timeout(command_timeout) do
+        openstack_command *args
+      end
+    rescue Timeout::Error
+      raise Puppet::ExecutionFailure, "Command: 'openstack #{args.inspect}' has been running for more then #{command_timeout} seconds!"
+    end
+  end
+
+  # get the current timestamp
+  def self.current_time
+    Time.now.to_i
+  end
 
   # Returns an array of hashes, where the keys are the downcased CSV headers
   # with underscores instead of spaces
@@ -18,14 +61,15 @@ class Puppet::Provider::Openstack < Puppet::Provider
     env = credentials ? credentials.to_env : {}
     Puppet::Util.withenv(env) do
       rv = nil
-      timeout = 10
-      end_time = Time.now.to_i + timeout
+      end_time = current_time + request_timeout
       loop do
         begin
-          if(action == 'list')
+          if action == 'list'
+            # shell output is:
+            # ID,Name,Description,Enabled
             response = openstack(service, action, '--quiet', '--format', 'csv', properties)
             response = parse_csv(response)
-            keys = response.delete_at(0) # ID,Name,Description,Enabled
+            keys = response.delete_at(0)
             rv = response.collect do |line|
               hash = {}
               keys.each_index do |index|
@@ -34,12 +78,15 @@ class Puppet::Provider::Openstack < Puppet::Provider
               end
               hash
             end
-          elsif(action == 'show' || action == 'create')
+          elsif action == 'show' or action == 'create'
             rv = {}
-            # shell output is name="value"\nid="value2"\ndescription="value3" etc.
+            # shell output is:
+            # name="value1"
+            # id="value2"
+            # description="value3"
             openstack(service, action, '--format', 'shell', properties).split("\n").each do |line|
               # key is everything before the first "="
-              key, val = line.split("=", 2)
+              key, val = line.split('=', 2)
               next unless val # Ignore warnings
               # value is everything after the first "=", with leading and trailing double quotes stripped
               val = val.gsub(/\A"|"\Z/, '')
@@ -49,24 +96,13 @@ class Puppet::Provider::Openstack < Puppet::Provider
             rv = openstack(service, action, properties)
           end
           break
-        rescue Puppet::ExecutionFailure => e
-          if e.message =~ /HTTP 401/
-            raise(Puppet::Error::OpenstackUnauthorizedError, 'Could not authenticate.')
-          elsif e.message =~ /Unable to establish connection/
-            current_time = Time.now.to_i
-            if current_time > end_time
-              break
-            else
-              wait = end_time - current_time
-              Puppet::debug("Non-fatal error: \"#{e.message}\"; retrying for #{wait} more seconds.")
-              if wait > timeout - 2 # Only notice the first time
-                notice("#{service} service is unavailable. Will retry for up to #{wait} seconds.")
-              end
-            end
-            sleep(2)
-          else
-            raise e
-          end
+        rescue Puppet::ExecutionFailure => exception
+          raise Puppet::Error::OpenstackUnauthorizedError, 'Could not authenticate' if exception.message =~ /HTTP 40[13]/
+          raise exception if current_time > end_time
+          debug "Non-fatal error: '#{exception.message}'. Retrying for #{end_time - current_time} more seconds"
+          raise exception if no_retry_actions.include? action
+          sleep retry_sleep
+          retry
         end
       end
       return rv
